
/*
 * $Id: CltXact.cc,v 1.25.2.6 1999/10/23 23:11:36 rousskov Exp $
 *
 *
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1999 by Alex Rousskov and the
 *  University of California, San Diego (UCSD); all rights reserved.
 *
 *  You may copy and distribute VERBATIM copies of this product. 
 *  You may modify your copy for your own use. You must NOT
 *  distribute modified copies. You must NOT publish benchmarking
 *  results based on modified versions.
 *
 *  If you publish any materials derived from or related to Web
 *  Polygraph, you must give a credit reference to the IRCache group.
 *
 *  Web Polygraph is distributed without any warranty, even implied.
 *
 *  Please see the COPYRIGHT file for full details.
 *
 */

#include "polygraph.h"

#include <ctype.h>
#include <strstream.h>
#include <iomanip.h>

#include "Clock.h"
#include "Connection.h"
#include "StatPhaseMgr.h"
#include "SharedOpts.h"
#include "ConnMgr.h"
#include "ErrorMgr.h"
#include "CltOpts.h"
#include "Client.h"
#include "ServerRep.h"
#include "CltXact.h"

#include "polyErrors.h"
#include "gadgets.h"
#include "globals.h"


/* HTTP header strings (to avoid strlen() calls when creating headers) */

// request line parts
static const String rlpGetHttpOrg = "GET /";
static const String rlpGetHttpPxy = "GET http://";
static const String rlpHttp_10    = " HTTP/1.0\r\n";

// full header fields
static const String hfAccept       = "Accept: */*\r\n";
static const String hfConnAliveOrg = "Connection: keep-alive\r\n";
static const String hfConnCloseOrg = "Connection: close\r\n";
static const String hfConnAlivePxy = "Proxy-Connection: keep-alive\r\n";
static const String hfConnClosePxy = "Proxy-Connection: close\r\n";

// header field prefixes
static const String hfpHost         = "Host: ";
static const String hfpXOidShortage = "X-Oid-Shortage: ";
static const String hfpXXact        = "X-Xact: ";


void CltXact::reset() {
	Xaction::reset();
	theOwner = 0;
	theConsumedSize = 0;
	theState = stNone;
	theSrvRep = 0;
	theOidShortage = 0;
	theRepHdr.reset();
}

Agent *CltXact::owner() {
	return theOwner;
}

void CltXact::enqueue() {
	theEnqueTime = TheClock;
}

void CltXact::exec(Client *anOwner, Connection *aConn) {
	Assert(theOidRec.wid);
	Assert(anOwner);
	theOwner = anOwner;

	start(aConn);

	theId.create();
	theSrvRep = Client::GetServerRep(theOidRec.sidx);

	makeReq(theConn->theWrBuf);
	theState = stSpaceWaiting;
	theConn->theWr.start(this);
}


void CltXact::finish(int errNo) {
	// get rid of reporting debt
	if (theOidShortage > 0)
		theSrvRep->noteNewOids(-1, theOidShortage, 0);

	stop(errNo);
	theOwner->noteXactDone(this);
}

void CltXact::noteReadReady(int) {
	theConn->read();

	if (theConn->bad()) {
		finish(-1);
		return;
	}

	if (theState == stHdrWaiting) {
		if (theRepHdr.parse(theConn->theRdBuf.content(), theConn->theRdBuf.contSize())) {

			// dump reply header
			if (TheOpts.theDumpFlags(dumpRep, dumpHdr))
				printMsg(theConn->theRdBuf, theRepHdr.theHdrSize);

			if (const int errNo = interpretHeader()) {
				finish(errNo);
				return;
			}

			consume(theRepHdr.theHdrSize);
			// fall through
		} else
		if (theConn->theRdBuf.full()) { // too big header
			finish(errHugeHdr);
			return;
		} else
		if (theConn->atEof()) { // premature end of headers
			finish(errPrematureEoh);
			return;
		}
	}

	if (theState == stBodyWaiting) {
		const Size parsedSize = parseContent();
		consume(parsedSize);

		if (theConsumedSize == theRepSize) { // got all we need
			if (theOidRec.foreign)
				finish(-1); // see errForeignRep elsewhere
			else
			if (theConn->theRdBuf.contSize() && !theConn->reusable())
				finish(errExtraRepData);
			else
				finish(0);
			return;
		}

		// tag without a closing bracket?
		if (theConn->theRdBuf.contSize() && theConn->atEof()) {
			finish(errOpenTag);
			return;
		}
	}

	if (theConn->exhausted()) { // premature eof
		finish(errPrematureEof);
		return;
	}
}

void CltXact::noteWriteReady(int) {

	Assert(theConn->theWrBuf.contSize() > 0);
	theConn->write();

	if (theConn->bad()) {
		finish(-1);
		return;
	}

	if (theConn->theWrBuf.empty()) {
		theConn->theWr.stop(this);
		theState = stHdrWaiting;
		theConn->theRd.start(this);
	} else
	if (!theConn->theWr.theReserv)
		theConn->theWr.start(this);
}

void CltXact::makeReq(WrBuf &buf) {
	const char *reqStart = buf.space();
	ostrstream os(buf.space(), buf.spaceSize());

	const NetAddr &orig = theSrvRep->addr();

	/* request line */

	// Method and Request-URI prefix (including first '/' in the path)
	if (TheCltOpts.thePxyAddr) {
		os << rlpGetHttpPxy << orig.addrA();
		if (orig.port() != 80)
			os << ':' << orig.port();
		os << '/';
	} else {
		// transparent folks get an advantage of shorter request-lines
		os << rlpGetHttpOrg;
	}

	theOidRec.printUriPath(os);
	os << rlpHttp_10; // end-of-request-line

	/*
	 * header fields: it is "good practice" to send general-header
	 * fields first, followed by request-header or response- header
	 * fields, and ending with the entity-header fields.
	 */

	/* general-header fields */

	// indicate persistency (HTTP/1.0 defaults to non-persistent)
	if (theConn->reusable())
		os << (TheCltOpts.thePxyAddr ? hfConnAlivePxy : hfConnAliveOrg);

	// request-header fields
	os
		<< hfAccept
		<< hfpHost << orig.addrA() << ':' << orig.port() << crlf
		;

	// entity-header fields
	os << hfpXXact << theId << crlf;
	
	if ((theOidShortage = theSrvRep->oidShortage()))
		os << hfpXOidShortage << theOidShortage << crlf;

	os << crlf; // end-of-header

	buf.appended(os.tellp());

	// dump request header
	if (TheOpts.theDumpFlags(dumpReq, dumpHdr))
		printMsg(reqStart, buf.space()-reqStart);
}

void CltXact::consume(Size size) {
	theConn->theRdBuf.consumed(size);
	theConsumedSize += size;
}

int CltXact::interpretHeader() {

	if (theRepHdr.theContSize < 0) // no content len
		return errNoCLen;

	theRepSize = theRepHdr.theHdrSize + theRepHdr.theContSize;

	// note: read foreign replies, but
	// do not count foreign replies as misses, hits, etc.!
	if (theRepHdr.domestic()) {
		theOidRec.cachable = theRepHdr.isCachable;
		theOidRec.hit = !theId.myMutant(theRepHdr.theXactId);

		if (!theOidRec.repeat && theOidRec.hit && !TheCltOpts.ignoreFalseHits)
			return errFalseHit;

		if (!theOidRec.cachable && theOidRec.hit)
			return errUnchbHit;

		// sync oid accounting if response headers are fresh (not a hit)
		if (!theOidRec.hit) {
			theSrvRep->syncSrvInfo(theRepHdr.theSrvWid, theRepHdr.theSecOid);
			if (theRepHdr.theNewOid > 0 && theRepHdr.theNewOidCnt > 0) {
				theSrvRep->noteNewOids(theRepHdr.theNewOid, theOidShortage, theRepHdr.theNewOidCnt);
				theOidShortage = -1; // no reporting debt left
			}
		}

		if (TheCltOpts.printFalseMisses && theOidRec.repeat && theOidRec.cachable && !theOidRec.hit) {
			theOidRec.print(cout << "False-Miss: ") << endl;
			printMsg(theConn->theRdBuf, theRepHdr.theHdrSize);
		}
	} else {
		theOidRec.foreign = true;
		if (ReportError(errForeignRep))
			printMsg(theConn->theRdBuf);
	}
	
	// note if the server will close the connection
	theConn->lastUse(!theRepHdr.keepAlive());

	theState = stBodyWaiting;
	return 0;
}

// returns number of bytes that can be consumed
Size CltXact::parseContent() {
	const IOBuf &buf = theConn->theRdBuf;
	const char *beg = buf.content();
	const char *end = beg + buf.contSize();
	const char *p = beg;

	// do not attempt to parse foreign content
	// do not attempt to parse embedded content
	if (theOidRec.foreign || theOidRec.embedded())
		return end - beg;

	while (p < end) {
		// find next tag opening
		if (const char *tagO = StrBoundChr(p, '<', end)) {
			p = tagO++;
			// find tag closing
			if (const char *tagC = StrBoundChr(tagO, '>', end)) {
				// found complete tag, parse it
				parseTag(tagO, tagC);
				p = tagC+1;
				// continue
			} else
				break; // wait for more data, leave tag prefix
		} else {
			p = end; // no more tags till the end of a buffer
			break;
		}
	}

	return p - beg;
}

void CltXact::parseTag(const char *tagB, const char *) {
	OidRec er;
	const char *p;
	if (isInt(tagB, er.tid, &p) && *p == '/' && isInt(p+1, er.oid, &p, 16)) {
		Assert(er.tid > 0);
		Assert(theOidRec.sidx >= 0);
		er.sidx = theOidRec.sidx;
		er.wid = theOidRec.wid;
		// note: .repeat assumes that objects have constant content
		// change this to (less precise) .hit when objects start changing
		er.repeat = theOidRec.repeat;
		theEmbedCount++;
		theOwner->noteEmbedded(this, er);
	} else {
		ReportError(isdigit(*tagB) ? errMalformedTag : errForeignTag);
	}
}
