
/*
 * $Id: PolyApp.cc,v 1.23.2.15 1999/11/08 06:38:37 rousskov Exp $
 *
 *
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1999 by Alex Rousskov and the
 *  University of California, San Diego (UCSD); all rights reserved.
 *
 *  You may copy and distribute VERBATIM copies of this product. 
 *  You may modify your copy for your own use. You must NOT
 *  distribute modified copies. You must NOT publish benchmarking
 *  results based on modified versions.
 *
 *  If you publish any materials derived from or related to Web
 *  Polygraph, you must give a credit reference to the IRCache group.
 *
 *  Web Polygraph is distributed without any warranty, even implied.
 *
 *  Please see the COPYRIGHT file for full details.
 *
 */

#include "polygraph.h"

#include <signal.h>

#include <fcntl.h>
#include <sys/stat.h>
#include <iostream.h>
#include <fstream.h>
#include <iomanip.h>

#include "Socket.h"
#include "AlarmClock.h"
#include "FileScanner.h"
#include "ResourceUsage.h"
#include "Rnd.h"
#include "RndPermut.h"
#include "HttpDate.h"
#include "StatCycle.h"
#include "StatPhase.h"
#include "StatPhaseMgr.h"
#include "StatsSampleMgr.h"
#include "ErrorStrTbl.h"
#include "ErrorMgr.h"
#include "Notifier.h"
#include "PolyOLog.h"
#include "LogComment.h"
#include "SharedOpts.h"
#include "Agent.h"
#include "UniqId.h"

#include "PglPp.h"
#include "PglParser.h"
#include "PglSemx.h"
#include "PglNetAddrRange.h"
#include "AgentSym.h"
#include "AgentArrIter.h"
#include "PhaseSym.h"

#include "PolyApp.h"

#include "gadgets.h"
#include "getIfAddrs.h"
#include "rndDistrs.h"
#include "polyLogTags.h"
#include "polyErrors.h"
#include "polyBcastChannels.h"
#include "shutdown.h"
#include "globals.h"

static StatCycle *TheStatCycle = 0;
static StatsSampleMgr TheStatsSampleMgr;

inline ostream &xexit(ostream &os) { exit(-2); return os; }

// make 'rw' permissions flag to use in open(2); move this?
#ifndef S_IRWMOST
#	if defined(S_IRUSR) && defined(S_IRGRP) & defined(S_IRWXO)
#		define S_IRWMOST (S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH)
#   else
#		define S_IRWMOST 0000666
#	endif
#endif


/* AgentCfg */

AgentCfg::AgentCfg(): sym(0) {
}

AgentCfg::AgentCfg(AgentSym *aSym, const NetAddr &aHost): sym(aSym) {
	host = new NetAddr(aHost);
}


/* PolyApp */

PolyApp::PolyApp(): agentMustBind(false), isIdle(false), theStateCount(0) {
	theChannels.append(TheShutdownChannel);
}

PolyApp::~PolyApp() {
	for (int i = 0; i < theLocals.count(); ++i)
		delete theLocals[i];
}

// scans all FDs once
void PolyApp::step() {
	Time tout = isIdle ? theIdleEnd - TheClock : (Time)TheOpts.theIdleTout;
	Time *toutp = &tout;

	if (TheAlarmClock.clear()) {
		if (tout < 0)
			toutp = 0;
	} else
	if (tout < 0) {
		tout = TheAlarmClock.timeLeft();
	} else {
		tout = Min(tout, TheAlarmClock.timeLeft());
	}

	const int readyCount = TheFileScanner->scan(toutp);

	if (ShutdownRequested())
		return;

	if (readyCount < 0 && errno != EINTR)
		FatalError(errno);

	// start "idle state" if needed
	if (!readyCount) {
		if (!isIdle && TheOpts.theIdleTout >= 0) {
			isIdle = true;
			theIdleBeg = TheClock;
			theIdleEnd = theIdleBeg + TheOpts.theIdleTout;
		}
	} else {
		isIdle = false;
	}

	Clock::Update();

	if (!readyCount && isIdle && TheClock.time() >= theIdleEnd)
		wasIdle(TheClock - theIdleBeg);
}

void PolyApp::begCycle(int) {
	Clock::Update(false);
	theTickCount = 0;
}

void PolyApp::endCycle() {
}

bool PolyApp::tick() {
	if (++theTickCount % 8 == 0)
		Clock::Update(false);
	return !ShutdownRequested();
}

void PolyApp::noteMsgStrEvent(BcastChannel *ch, const char *msg) {
	Assert(ch == TheShutdownChannel);
	ShutdownReason(msg);
}

void PolyApp::wasIdle(Time tout) {
	if (TheOpts.theIdleTout >= 0 && tout >= TheOpts.theIdleTout) {
		Comment << "was idle for at least " << tout << endc;
		ShutdownReason("inactivity timeout");
	}
}

bool PolyApp::handleCmdLine(int argc, char *argv[]) {

	Array<OptGrp*> opts;
	getOpts(opts);

	theCmdLine.configure(opts);
	if (!theCmdLine.parse(argc, argv))
		return false;

	// validate command line params
	for (int i = 0; i < opts.count(); ++i)
		if (!opts[i]->validate())
			return false;

	return true;
}

void PolyApp::configureBinLog(OLog &log, const String &fname, SizeOpt &bufSz) {
	const int fd = ::open(fname, O_WRONLY|O_CREAT|O_TRUNC, S_IRWMOST);
	if (fd < 0)
		cerr << ThePrgName << ": cannot create binary log; " 
			<< fname << ": " << error(errno) << endl << xexit;

	log.stream(fname, fd);

	// we cannot shrink log size
	if (bufSz > log.capacity())
		log.resize(bufSz); 
	bufSz.set(log.capacity());
}

void PolyApp::configureLogs(int prec) {
	// redirect console output
	if (TheOpts.theConsFileName && strcmp(TheOpts.theConsFileName, "-"))
		cout = cerr = clog = *new ofstream(TheOpts.theConsFileName);

	configureStream(cout, prec);
	configureStream(cerr, prec);
	configureStream(clog, prec);

	/* initialize binary log files */

	if (TheOpts.theLogFileName) {
		configureBinLog(TheOLog, TheOpts.theLogFileName, TheOpts.theLogSize);
		TheOLog.period(Time::Sec(5)); // flush at most that often
	}
	
	if (TheOpts.theSmplLogFileName) {
		TheSmplOLog = new PolyOLog();
		configureBinLog(*TheSmplOLog, TheOpts.theSmplLogFileName, TheOpts.theSmplLogSize);
		// sample log is not flushed on a periodic basis
	} else {
		// use general purpose log
		TheOpts.theSmplLogFileName.val(TheOpts.theLogFileName);
		TheOpts.theSmplLogSize.set(TheOpts.theLogSize);
		TheSmplOLog = &TheOLog;
	}

	Comment.TheEchoLevel = TheOpts.theVerbLevel;
}

// get a list of all available interface addresses
void PolyApp::getIfaces() {
	// manually configured [faked] addresses take priority
	if (TheOpts.theFakeHosts) {
		getFakeIfaces();
	} else
	if (!::GetIfAddrs(theIfaces))
		cerr << ThePrgName << ": cannot get a list of all available network interfaces: " << error(errno) << endl << xexit;
}

void PolyApp::getFakeIfaces() {
	// note each list item may be an IP range
	for (int i = 0; i < TheOpts.theFakeHosts.val().count(); ++i) {
		const String &item = *TheOpts.theFakeHosts.val()[i];
		PglNetAddrRange range;
		if (!range.parse(item))
			cerr << ThePrgName << ": cannot convert fake host `" << item 
				<< "' to an IP address or range" << endl << xexit;

		Array<NetAddr*> addrs;
		range.toAddrs(addrs);
		for (int a = 0; a < addrs.count(); ++a)
			theIfaces.append(addrs[a]->addrN());
		while (addrs.count()) delete addrs.pop();
	}
}

// used by countAddresses
int PolyApp::countAddress(const NetAddr &addr) const {
	int count = 0;
	for (int i = 0; i < theLocalCfgs.count(); ++i) {
		if (theLocalCfgs[i].host->addrN() == addr.addrN())
			count++;
	}
	return count;
}

bool PolyApp::moreThanOneAddr() const {
	const int count = theLocalCfgs.count();
	for (int i = 0; i < count; ++i) {
		if (countAddress(*theLocalCfgs[i].host) < count)
			return true;
	}
	return false;
}

void PolyApp::makeAgents() {
	// PglCtx::RootCtx()->report(cout, "");

	getIfaces();

	// collect agents that are assigned to our host
	for (AgentArrIter i(PglSemx::TheAgentsToUse, &theIfaces, theAgentType); i; ++i) {
		theLocalCfgs.append(AgentCfg(i.agentSym(), i.host()));
	}

	agentMustBind = moreThanOneAddr();

	// create and configure agents
	const Time cfgReportGap = Time::Sec(5); // how often to report
	const Time cfgStart = TheClock;
	Time cfgNextReport = cfgStart + cfgReportGap;
	for (int i = 0; i < theLocalCfgs.count(); ++i) {
		Agent *agent = makeAgent(theLocalCfgs[i]);
		Assert(agent);
		theLocals.append(agent);

		// configuration process may take a while; inform about the progress
		Clock::Update(false);
		if (TheClock.time() >= cfgNextReport) {
			Comment(5) << "created " << setw(6) << i+1 << " out of " 
				<< theLocalCfgs.count() << " agents so far" << endc;
			cfgNextReport = TheClock + cfgReportGap; // drift is OK
		}
	}
	Clock::Update(false);

	if (!theLocals.count()) {
		cerr << ThePrgName << ": no " << theAgentType << " matches local interface addresses" << endl;

		cerr << theAgentType << " addresses:";
		for (AgentArrIter i(PglSemx::TheAgentsToUse, 0, theAgentType); i; ++i)
			cerr << ' ' << i.host();
		cerr << endl;

		cerr << "local addresses: ";
		for (int i = 0; i < theIfaces.count(); ++i)
			cerr << ' ' << inet_ntoa(theIfaces[i]);
		cerr << endl;

		cerr << xexit;
	}
}

// configure StatPhaseMgr with user specified phases
void PolyApp::buildSchedule() {
	if (!PglSemx::TheSchedule.count()) {
		cerr << ThePrgName << ": warning: no run phases were specified; generating and using default phase" << endl;
		PhaseSym *ps = new PhaseSym();
		ps->loc(TokenLoc("-"));
		ps->name("dflt");
		PglSemx::TheSchedule.append(ps);
	}

	const StatPhase *prevPh = 0;
	for (int i = 0; i < PglSemx::TheSchedule.count(); ++i) {
		StatPhase *p = new StatPhase;
		p->configure(PglSemx::TheSchedule[i], prevPh);
		TheStatPhaseMgr.addPhase(p);
		prevPh = p;
	}
}

// top level configure routine
void PolyApp::configure() {

	configureLogs(2);

	startListen();

	TheErrorMgr.configure(polyErrorStrTbl());

	// ignore some signals
	signal(SIGPIPE, SIG_IGN);
	signal(SIGHUP, SIG_IGN);

	// set resource limits
	const int rlimit = SetMaxFD(GetMaxFD());
	if (rlimit > FD_SETSIZE)
		cerr << here << "warning: getrlimit(2) system call reports " << rlimit << " file descriptors while FD_SETSIZE #defines only " << FD_SETSIZE << "; using lower value." << endl;
	const int fdLimit = Min(FD_SETSIZE, rlimit);
	if (fdLimit <= 0)
		FatalError(errZeroFDLimit);
	TheFileScanner = TheOpts.theFileScanner.val();
	TheFileScanner->configure(fdLimit);
	TheFileScanner->ticker(this);

	// to be safe, cut about 3% from the actual limit
	Socket::TheMaxLevel = Max(0, fdLimit - 10 - fdLimit/33);
	// honor manually configured limit if it is lower
	if (TheOpts.theFDLimit >= 0 && TheOpts.theFDLimit < Socket::TheMaxLevel)
		Socket::TheMaxLevel = TheOpts.theFDLimit;
	TheOpts.theFDLimit.set(Socket::TheMaxLevel);

	// set random seeds
	RndPermutSeed(TheOpts.theRngSeed);
	RndGen::TheDefSeed = RndPermut(TheOpts.theRngSeed);

	/* statistics */

	TheStatCycle = new StatCycle;
	TheStatCycle->period(TheOpts.theStatCycleLen);
	buildSchedule(); // configures TheStatPhaseMgr

	if (PglSemx::TheSmplSchedule.count()) {
		TheStatsSampleMgr.configure(PglSemx::TheSmplSchedule);
		TheStatsSampleMgr.start();
	}

	// configure notification mechanism
	if (TheOpts.theNotifAddr) {
		Notifier *n = new Notifier(TheOpts.theRunLabel, TheOpts.theNotifAddr);
		if (n->active())
			TheStatCycle->notifier(n);
		else
			delete n;
	}

	atexit(&Shutdown);
}

void PolyApp::getOpts(Array<OptGrp*> &opts) {
	opts.append(&TheOpts);
}

void PolyApp::parseConfigFile(const String &fname) {
	Assert(fname);
	TheOpts.theCfgDirs.copy(PglPp::TheDirs);

	PglPp pp(fname);
	PglParser parser(&pp);

	parser.parse();
}

void PolyApp::reportRUsage() const {
	ResourceUsage ru = ResourceUsage::Current();
	Comment(5) << "resource usage: " << endl;
	ru.report(Comment, "\t");
	Comment << endc;
}

void PolyApp::reportCfg() {

	theCmdLine.reportRaw(Comment(1)); Comment << endc;
	theCmdLine.reportParsed(Comment(2)); Comment << endc;

	TheStatPhaseMgr.reportCfg(Comment(2) << "Phases:" << endl);
	Comment << endc;
	TheStatsSampleMgr.reportCfg(Comment(2) << "StatsSamples:" << endl);
	Comment << endc;

	Comment(2) << "FDs: "
		<< GetCurFD() << " out of " << GetMaxFD()
		<< " FDs can be used; safeguard limit: " << Socket::TheMaxLevel 
		<< endc;

	reportRUsage();

	const UniqId runId = UniqId::Create();
	Comment(1) << "run-id: " << runId << " pid: " << getpid() << endc;
	TheOLog << bege(lgRunId) << runId << ende;

	Comment << "current time: " << TheClock.time() << " or ";
	HttpDatePrint(Comment, TheClock.time()) << endc;
	Comment << "configuring agents and waiting for traffic" << endc;
}

void PolyApp::flushState() {
	theStateCount++;
	logState(TheOLog);
	Comment(5) << "fyi: current state (" << theStateCount << ") stored" << endc;
}

void PolyApp::logState(OLog &log) {
	TheOLog << bege(lgAppState) << theStateCount << ende;
	Broadcast(TheLogStateChannel, &log);
}

int PolyApp::run(int argc, char *argv[]) {

	ThePrgName = argv[0];

	Clock::Update();

	if (!handleCmdLine(argc, argv))
		return -1;

	parseConfigFile(TheOpts.theCfgFileName);

	Clock::Update();

	configure();
	makeAgents();

	reportCfg();

	// store errno strings (should be in logCfg)
	TheOLog << bege(lgErrorStrs) << *polyErrorStrTbl() << ende;

	flushState();

	// flush logs headers
	TheOLog.flush(); 
	TheSmplOLog->flush();

	// quit nicely on some signals
	// note: install this handler after configuration is done
	// so a program can be killed if configure takes forever
	signal(SIGINT, (SignalHandler*)&ShutdownSignal);

	Clock::Update();

	TheStatPhaseMgr.start();
	TheStatCycle->start(); 

	Clock::Update();

	for (int i = 0; i < theLocals.count(); ++i)
		theLocals[i]->start();

	Clock::Update();

	while (!ShutdownNow()) {
		step();
	}

	reportRUsage();
	flushState();

	return 0;
}


#if 0 /* main() template */
int main(int argc, char *argv[]) {

	XxxApp app;
	return app.run(argc, argv);
}
#endif
